//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: agg_agg.C,v 1.32 2004/11/16 19:48:59 adams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           agg_agg.C
//   
#include "prometheus.hh" 
// module data (hack) -- non-rentrant
#define ADJAC_FACTOR 8
#define EMAX 12800
static  idxtype *xadj = NULL;
static  int edgeW_limit = -1, *ewgts = NULL, numadjac = 0; 
static  int hit, miss, global_agg = 1;

#undef __FUNCT__    
#define __FUNCT__ "PromAggGrid::Coarsen" 
/* PromAggGrid::Coarsen *******************************************************
 *
 *  - take this mesh and coarsen with partitioner
 * 
 *  INPUT:
 *
 *  SIDE EFFECTS:
 *    - constructs selected child lists
 *
 *  RETURN:
 *    - PET cerror code 
 */
int PromAggGrid::Coarsen()
{
  int ierr,ii,vb=options_.verbose_;   assert(isActive());
  bool new_agg,use_metis;

  PetscFunctionBegin;
  //edgeW_limit = 4*(EMAX/100); // Vanek: .08 * l^0.5
  edgeW_limit = 8*(EMAX/100);   // global param
  for(ii=0;ii<getLevel();ii++) edgeW_limit = edgeW_limit / 2;

  // construct initail partitions
  ierr = options_.HasOptionName("-aggmg_use_new_agg",&new_agg); CHKERRQ(ierr);
  ierr = options_.HasOptionName("-aggmg_use_metis",&use_metis); CHKERRQ(ierr);
#if defined(PROM_HAVE_METIS)
  if( amg_factor_ == 1 || new_agg ) global_agg = 1;
  else global_agg = 0;
#else
  global_agg = 1; use_metis = false;
#endif

  // alloc global var!!!
  ierr = PetscMalloc((nLocalNd_+1)*sizeof(idxtype), &xadj);  CHKERRQ(ierr);
  if( new_agg ) ierr = PartitionAggs_new();
  else ierr = PartitionAggs();  CHKERRQ(ierr);

  if(vb>1)PetscPrintf( gComm_, 
		       "[%d]%s %s, %s agg. method, edgeW_limit=%g hit=%d miss=%d\n",
		       comm_.mype(),__FUNCT__, !use_metis ? 
		       "MIS Aggragates" : "Partition Aggragates",
		       new_agg ? "New" : "Old",
		       (float)edgeW_limit/(float)EMAX,hit,miss );

  // aggregate and complete
  ierr = ConstructAggs();  CHKERRQ(ierr); // hangs ?
  // clean up
  PetscFree( xadj );  xadj = NULL;  // static 
  if( ewgts != NULL ) PetscFree( ewgts ); ewgts = NULL; // static
  
  // output
  ierr = options_.HasOptionName( "-out_aggs", &new_agg );  CHKERRQ(ierr);
  if( new_agg || options_.verbose_ > 2) { // debug
    char root[64], name[64];
    ierr = options_.GetOptionString("-out_aggs", root, 63, &new_agg);
    CHKERRQ(ierr);
    if( !new_agg || strlen(root) == 0 ) sprintf( root, "aggregates" );
    sprintf( name, "%s_grid_%d", root, getLevel() );
    for( int p = 0, nn = 0 ; p < comm_.np() ; p += factor_ ){
      MPI_Allreduce( &nn, &ii, 1, MPI_INT, MPI_MAX, gComm_ );
      if( p != comm_.mype() ) nn = 0;
      else {
	nn = ii;
	FILE *file = fopen( name, (p == 0) ? "w" : "a" );
	if( !file ) SETERRQ(1,name); 
	for( int xx = 0 ; xx < nLocalNd_ ; xx++ ) {
	  PromNode *curr = nodes_[xx];
	  if( curr->getState() == PROM_SELECTED ){
	    for(PromNode *c1=curr ; c1 ; c1=(PromNode*)c1->getChild() ){
	      fprintf( file, "%d %d\n", nn, c1->index_.global() );
	    }
	    nn++;
	  }
	}
	fclose(file);
      }
    }
  }

  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "PromIsAMGDeleteable" 
/* PromIsAMGDeleteable *************************************************
 *
 *  - see if this can delete n1
 *  
 *  INPUT:
 *   - curr1: 'this' (symmetric on args)
 *   - curr2: node to check
 *   - killit: (out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */
static int PromIsAMGDeleteable( const PromNode * const curr1, 
				const PromNode * const curr2, bool &del)
{
  int lid, *whts;
  const PromNode *n2, *n1;

  if( curr1->getNDF() != curr2->getNDF() ){ del = FALSE;  return 0; }
  else del = TRUE;
  
  if( ewgts == NULL ) { del = TRUE;  return 0; }
   
  if( curr1->isGhost() ){ n1 = curr2; n2 = curr1; }
  else{ n1 = curr1; n2 = curr2; }
  
  const int limit = edgeW_limit; 
  lid = n1->index_(); 
  whts = &ewgts[xadj[lid]]; // start of weights
  PromTable<PromNode*> *adj = &n1->symbData_->adjac;
  TablePos tpos = adj->getHeadPosition();assert(tpos!=0);
  while( tpos ){ 
    PromNode *curr = adj->getNext(tpos);
    if( n1->getNDF() == curr->getNDF() ) {
      int wht = *whts++;
      if( curr == n2 ){
	if( wht > limit ){del = TRUE; hit++;} 
	else{ del = FALSE; miss++; }
	return 0;
      }
    }
  }
  SETERRQ(1,"N2 not found!!!");
}

#undef __FUNCT__    
#define __FUNCT__ "PromAggGrid::MakeWeightMatrix" 
/* PromAggGrid::MakeWeightMatrix ********************************************
 *
 *  - Make wight matrix from node->'adjac'
 *
 *  INPUT:
 *   - ndwts: node wights (out)
 *   - padj: edges, allocated (out)
 *
 *  SIDE EFFECTS:
 *   - allocates 'ewgts'
 *
 *  RETURN:
 *    - PETSc error code 
 */
int PromAggGrid::MakeWeightMatrix( int ndwts[], idxtype **padj )
{
  int xx,ierr,mm,ii,numadjacTOT=-1,*ewgtsTOT=NULL;
  PromNode *curr; idxtype *adjncy;

  // count edges for e weights ,set "xadj" 
  for( xx = ii = mm = 0 ; xx < nLocalNd_ ; xx++ ) {
    curr = nodes_[xx]; assert(curr->symbData_ != NULL);
    xadj[xx] = ii;  
    PromTable<PromNode*> *adj = &curr->symbData_->adjac;
    TablePos tpos = adj->getHeadPosition(); 
    while( tpos ){ 
      PromNode *n1 = adj->getNext(tpos); assert(n1!=curr);
      if( n1->getNDF() == curr->getNDF() ) {
	mm++; // total edeges
	if( global_agg || !n1->isGhost() ) ii++;
      }
    }
  }
  xadj[xx] = ii; assert(xx==nLocalNd_); 
  if(!global_agg && ii%2 != 0)SETERRQ1(1,"ii%2!=0 = %d",ii); 
  if ( global_agg ) numadjac = mm; 
  else numadjac = ii; 

  numadjacTOT = mm;  assert(numadjacTOT>=numadjac); 
  ierr = PetscMalloc( (numadjac+1)*sizeof(int), &ewgts ); CHKERRQ(ierr);
  ierr = PetscMalloc( (numadjac+1)*sizeof(idxtype), &adjncy ); CHKERRQ(ierr);
  if ( global_agg ) ewgtsTOT = ewgts;
  else {
    ierr = PetscMalloc((numadjacTOT+1)*sizeof(int),&ewgtsTOT); CHKERRQ(ierr);
  }
  // edge weights 
  assert(stiffness_ != NULL);
  ierr = SetAMGEdgeWeights( stiffness_, adjncy, numadjacTOT, ewgtsTOT );
  CHKERRQ(ierr);
  if ( !global_agg ) PetscFree( ewgtsTOT );
  // set node wieghts with: inverse of degree or type
  for( xx = 0 ; xx < nLocalNd_ ; xx++ ){
    curr = nodes_[xx];     
    PromTable<PromNode*> *adj = &curr->symbData_->adjac;
    ndwts[xx] = 110/adj->getCount() + 1;
  }
  *padj = adjncy;

  return 0;
}
#undef __FUNCT__    
#define __FUNCT__ "PromAggGrid::PartitionAggs" 
/* PromAggGrid::PartitionAggs ******************************************
 *
 *  - 
 *
 *  INPUT:
 *
 *  SIDE EFFECTS:
 *    - constructs selected child lists 
 *    - sets "numadjac", "ewgts"
 *
 *  RETURN:
 *    - PET cerror code 
 */
int PromAggGrid::PartitionAggs()
{
  int ierr,xx;
  int *ndwts=NULL, vb = options_.verbose_; idxtype *adjncy=NULL;
  PromNode *curr;
  bool mis2;
  int *origadjs = NULL;

  PetscFunctionBegin;
  // check for MIS on squared graph
  ierr = options_.GetOptionInt("-prometheus_mis_levels",&xx, &mis2);
  if( !mis2 || xx < 1 ) xx = 1;
  
  if( xx > 1 && isBase() ) {
    ierr = SquareGraph( &origadjs ); CHKERRQ(ierr);
  }
  else {
    // make edge wieghts (avoid if MIS^2)
    mis2 = FALSE;
    /* construct METIS vertex graph data structures */
    ierr = PetscMalloc((nLocalNd_+1)*sizeof(int), &ndwts ); CHKERRQ(ierr);
    //
    ierr = MakeWeightMatrix( ndwts, &adjncy ); CHKERRQ(ierr);
  }

  // doit
  if( global_agg ) {  
    // make node list for MIS (put boundaries at beginning)
    bool ranMIS;
    ierr = options_.HasOptionName("-prometheus_random_mis", &ranMIS);
    CHKERRQ(ierr);
    // make lists
    PromList<PromNode*> nodeList;
    if( ranMIS ) {
      PromTable<PromNode*> table(4*nLocalNd_/3+5); int cc;
      for(xx = cc = 0 ; xx < nLocalNd_ ; xx++) {
	curr = nodes_[xx];
	table.add( curr );
	cc += curr->symbData_->adjac.getCount();
      }
      cc = cc / nLocalNd_ + 1;
      while( !table.isEmpty() ) { 
	curr = table.removeHead();
	if( curr->symbData_->adjac.getCount() > cc) nodeList.addHead( curr );
	else nodeList.addTail( curr );
      }
    }
    else {
      for(xx = 0 ; xx < nLocalNd_ ; xx++) nodeList.addTail(nodes_[xx]);
    }

    // MIS
    hit = miss = 0; // global vars set in callback "PromIsAMGDeleteable"
    ierr = CreateMIS(nodeList,PromIsAMGDeleteable,FALSE); CHKERRQ(ierr);
    //isbase = FALSE; // just done once -- hack!
    if( origadjs != NULL ) {
      ierr = UnSquareGraph( &origadjs ); CHKERRQ(ierr);
    }
    // set children
    for( xx = 0 ; xx < nNodes_ ; xx++ ) {
      curr = nodes_[xx];
      if( curr->getState() == PROM_DELETED ) {
	PromNode *sel = curr->symbData_->parent; 
	ierr = sel->addChild( curr );
      }
    }
  }
  else {
    // get num domains
    int sumWt,nbglobal = (int)((float)nglobal()/(float)amg_factor_+.5);
    for(xx=sumWt=0;xx<numadjac;xx++)sumWt += ewgts[xx];
    MPI_Allreduce( &sumWt, &xx, 1, MPI_INT, MPI_SUM, gComm_ );
    int nbloc = (int)((float)nbglobal*(float)sumWt/(float)xx+0.5); // my share
    assert(nbloc>=0);
    if( nbloc <= 0 ) nbloc = 1;
    if( nLocalNd_ < 5 ) nbloc = 1;
    MPI_Allreduce( &nbloc, &nbglobal, 1, MPI_INT, MPI_SUM, gComm_ );
    if(vb>1)PetscPrintf(vb > 2 ? PETSC_COMM_SELF : gComm_,
			"\t[%d]%s %d aggs. (%d local) amg factor=%d sumWt=%d\n",
			comm_.mype(),__FUNCT__,nbglobal,nbloc,amg_factor_,sumWt);
#if defined(PROM_HAVE_METIS)
    int *pnode_part;  
    ierr = PetscMalloc((nLocalNd_+1)*sizeof(int), &pnode_part); CHKERRQ(ierr);
    if( nbloc == 1 ) for( xx = 0 ; xx < nLocalNd_ ; xx++ ) pnode_part[xx] = 0;
    else {
      // 'ewgts' causes bug?
      int wt=(ndwts ? 2 : 0)+(ewgts ? 1 : 0);
      //int wt=(ndwts ? 2 : 0);
      int options[5]={0,3,1,1,0},nmf=0,edgecut=-1,nloc=nLocalNd_;;
      perf_mon_.EventBegin( PromContext::prom_logs_[LOG_METIS] );
      METIS_PartGraphKway( &nloc, xadj, adjncy, ndwts, ewgts, &wt, &nmf,
			   &nbloc, options, &edgecut, pnode_part );
      perf_mon_.EventEnd(PromContext::prom_logs_[LOG_METIS] );  
    }
    //
    ierr = PickSubdomains( pnode_part );        CHKERRQ(ierr);
    PetscFree( pnode_part ); 
#else
    SETERRQ(1,"Aggregation 'factor' not supported for build w/o METIS");
#endif
  }
  // clean up
  if( ndwts != NULL ) PetscFree( ndwts );
  if( adjncy != NULL ) PetscFree( adjncy );

  PetscFunctionReturn(0);
}

static int get_norm( const int indf, const int fact, const int gidj, 
		     const int jndf, const int ncols, const int adjacs[], 
		     const double * const Ai[], double *ret ) 
{
  int j,i;
  double mx,tt;

  // linear search!!!
  for(i=0, j=-1 ; i < ncols && j == - 1 ; i++)if(adjacs[i] == gidj) j = i;
  if(j == -1) {
    *ret = 0.; // from skip smoothing
    return 1;
  }
  else {
    const double *aa, *pAij = Ai[j]; // assumes blocks stored transpose
    for( j = 0, mx = 0., aa = pAij ; j < jndf ; j++ ){
      for( i = 0 ; i < indf; i++, aa += fact ) {
	tt = fabs(*aa); if( tt > mx ) mx = tt;
      }
      if(fact==2) aa += 2*indf;
    }
    *ret = mx;
  }

  return 0;
} 

static int get_weight( const double naij, const double naii, 
		       const double najj, int *ret )
{
  int tt;
  double temp = naij / sqrt(naii*najj+1.e-200);
  if( naij > 0.0 && temp > 1.0 ){
    if( temp > 2.0 ) {
      PetscPrintf(PETSC_COMM_SELF,"[?]get_weight WARNING:|Aij^2|/|AiiAjj|=%g Aij=%g Aii=%g Ajj=%g dem=%g\n",temp,naij,naii,najj,sqrt(naii*najj+1.e-200));
    }
    tt = EMAX;
  }
  else if( temp <= 0.0 ) tt = 1; // fixed bc
  else {
    tt = (int)( (double)EMAX * temp + .5 );
    if( tt <= 0 ) tt = 1;	    
  }
  *ret = tt; 
  
  return 0;
}

static PromMatrix *mat_loc = NULL;
static int my0;
static int next0;

#undef __FUNCT__  
#define __FUNCT__ "PromIsAMGDeleteable_new" 
/* PromIsAMGDeleteable_new ********************************************
 *
 *  - see if this can delete n1
 *  
 *  INPUT:
 *   - curr1: 'this' (symmetric on args)
 *   - curr2: node to check
 *   - killit: (out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */
static int PromIsAMGDeleteable_new( const int curr1, const int curr2,
				    bool &del )
{
  int ii,jj,wht,ncols,*adjacs,*whts,ierr,gidi,gidj;

  del = TRUE;

  if( mat_loc == NULL ) { del = TRUE; return 1; }

  if( curr1 < my0 || curr1 >= next0 ){ gidi = curr2; gidj = curr1; }
  else{ gidi = curr1; gidj = curr2; }
  
  const int limit = edgeW_limit, lidi = gidi - my0;
  whts = &ewgts[xadj[lidi]]; // start of weights
  ierr = mat_loc->GetRow( gidi, &ncols, &adjacs );  CHKERRQ(ierr);
  for( ii = jj = 0, wht = -1 ; ii < ncols && wht == -1 ; ii++ ) {
    if( adjacs[ii] == gidj ) wht = whts[jj]; // break
    else if( adjacs[ii] != gidi ) jj++;
  }
  ierr = mat_loc->RestoreRow( gidi, &ncols, &adjacs ); CHKERRQ(ierr);

  if( wht > limit ){del = TRUE; hit++;} 
  else{ del = FALSE; miss++; }
  
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromAggGrid::MakeWeightMatrix" 
/* PromAggGrid::MakeWeightMatrix ********************************************
 *
 *  - Make wight matrix from 'stiffness_'
 *
 *  INPUT:
 *   - ndwts: node wights (out)
 *   - padj: edges, allocated (out)
 *   - mis2: two level MIS?
 *
 *  SIDE EFFECTS:
 *   - allocates 'mat_loc', 'ewgts'
 *
 *  RETURN:
 *    - PETSc error code 
 */
int PromAggGrid::MakeWeightMatrix( const PromMap &map, int ndwts[], 
				   idxtype **padj, bool mis2 /*=false*/)
{
  int ierr,xx,gidi,ii,ncols,adjacs[PROM_MAXNIEGH];
  const PromMatrix *const AA = stiffness_;
  const int fact = AA->isComplex() ? 2 : 1, nloc = nlocal();
  PromNode *curr;  double val,*Ai[PROM_MAXNIEGH];

  PetscFunctionBegin;
  // allocate (global) matrix object
  assert(mat_loc == NULL);
  mat_loc = new PromMatrix( map, 27, NULL, 9, NULL ); // global var!!

  for( xx = numadjac = 0, gidi = my0 ; xx < nloc ; xx++, gidi++ ) {
    curr = nodes_[xx];    const int indf = curr->getNDF();
    ierr = AA->GetLocalNodeRow( xx, PROM_MAXNIEGH, ncols, adjacs, Ai );
    CHKERRQ(ierr);
    int norms_i[PROM_MAXNIEGH]; double norms[PROM_MAXNIEGH];
    TablePos tpos = curr->symbData_->adjac.getHeadPosition(); ii = 0;
    while( tpos ) {
      PromNode *n1 = curr->symbData_->adjac.getNext(tpos);
      if( n1->getNDF() == curr->getNDF() ){
	int gidj = n1->index_.global();
	ierr = get_norm(indf, fact, gidj, indf, ncols, adjacs, Ai, &val);
	if(ierr)SETERRQ2(1,"get_norm: %d -- %d, not found",gidi,gidj);
	if( val < 0. ) val = -val;
	norms_i[ii] = gidj; norms[ii] = val; ii++;
      }
    }
    // add diagonal
    for( int jj = 0 ; jj < ncols ; jj++ ) {
      if( adjacs[jj] == gidi ){
	ierr = get_norm(indf, fact, gidi, indf, ncols, adjacs, Ai, &val);
	CHKERRQ(ierr);
	norms_i[ii] = gidi; norms[ii] = val; ii++;
      }
    }
    ierr = mat_loc->SetValues(1, &gidi, ii, norms_i, norms,INSERT_VALUES);
    CHKERRQ(ierr);
    xadj[xx] = numadjac;  numadjac += (ii-1); // diag
  }
  xadj[xx] = numadjac;  
  ierr = mat_loc->Assembly(); CHKERRQ(ierr);

  if( mis2 ) {
    PromMatrix *mat2;
    ierr = mat_loc->SetOption( MAT_STRUCTURALLY_SYMMETRIC );  CHKERRQ(ierr);
    ierr = mat_loc->Mult( mat_loc, &mat2 ); CHKERRQ(ierr);
    delete mat_loc; mat_loc = mat2;
    // redo xadj
    for( xx = numadjac = 0, gidi = my0 ; xx < nloc ; xx++, gidi++ ) {
      ierr = mat_loc->GetRow( gidi, &ii );         CHKERRQ(ierr);
      xadj[xx] = numadjac;  numadjac += (ii-1); // diag
      ierr = mat_loc->RestoreRow( gidi, &ii );  CHKERRQ(ierr);
    }
    xadj[xx] = numadjac;
  }
  ierr = mat_loc->SetOption(MAT_STRUCTURALLY_SYMMETRIC); CHKERRQ(ierr);
  { // normalize
    PromVector vec(map);
    ierr = mat_loc->GetDiagonal( &vec );      CHKERRQ(ierr);
    ierr = vec.Sqrt();                   CHKERRQ(ierr);
    ierr = vec.Reciprocal();             CHKERRQ(ierr);
    ierr = mat_loc->Scale( &vec, &vec );      CHKERRQ(ierr);
  }
  // make 'adjncy'
  idxtype *adjncy = NULL;
  ierr = PetscMalloc((numadjac+1)*sizeof(idxtype),&adjncy);
  CHKERRQ(ierr);      adjncy[numadjac] = -1;
  ierr = PetscMalloc((numadjac+1)*sizeof(idxtype),&ewgts); 
  CHKERRQ(ierr);      ewgts[numadjac] = -1;  
  for( xx = ii = 0, gidi = my0 ; xx < nloc ; xx++, gidi++ ) {
    int *adjacs; double *vv;
    ierr = mat_loc->GetRow( gidi, &ncols, &adjacs, &vv ); CHKERRQ(ierr);
    for( int i = 0 ; i < ncols ; i++ ) {
      if( adjacs[i] != gidi ){
	adjncy[ii] = adjacs[i];
	ewgts[ii] = 1 + (int)(EMAX*vv[i]); ii++;
      }
    }
    ndwts[xx] = 110/ncols + 1; 
    ierr = mat_loc->RestoreRow(gidi, &ncols,&adjacs,&vv); CHKERRQ(ierr);
  }
  assert(ii == numadjac);
  *padj = adjncy;

  PetscFunctionReturn(0);
}

#undef __FUNCT__    
#define __FUNCT__ "PromAggGrid::PartitionAggs_new" 
/* PromAggGrid::PartitionAggs_new ********************************************
 *
 *  - 
 *
 *  INPUT:
 *
 *  SIDE EFFECTS:
 *    - constructs selected child lists 
 *    - sets "numadjac", "ewgts"
 *
 *  RETURN:
 *    - PETSc error code 
 */
  int PromAggGrid::PartitionAggs_new()
{
  int ierr,xx,*ndwts=NULL,vb=options_.verbose_;   idxtype *adjncy = NULL;
  const int nloc=nlocal(); bool mis2;
  PromComm wcomm( comm_() ); PromMap map( nloc, 1, wcomm ); 
  PromNode *curr; 

  PetscFunctionBegin;  
  my0 = my0Nd(); next0 = my0 + nloc;                  // global var!

  // check for MIS on squared graph
  ierr = options_.GetOptionInt("-prometheus_mis_levels",&xx, &mis2);
  CHKERRQ(ierr);
  if( !mis2 || xx < 1 ) xx = 1;
  if( !isBase() || xx < 2 ) mis2 = false;

  ierr = PetscMalloc( (nloc+1)*sizeof(idxtype), &ndwts );  CHKERRQ(ierr);

  // make wieght matrix
  ierr = MakeWeightMatrix( map, ndwts, &adjncy, mis2 ); CHKERRQ(ierr);

  // doit
  if( global_agg ) {  
    // make node list for MIS (put boundaries at beginning)
    bool ranMIS;
    ierr = options_.HasOptionName("-prometheus_random_mis", &ranMIS);
    CHKERRQ(ierr);
    // make lists
    PromList<PromNode*> nodeList;
    if( ranMIS ) {
      PromTable<PromNode*> table(4*nLocalNd_/3+5); int cc;
      for(xx = cc = 0 ; xx < nLocalNd_ ; xx++) {
	curr = nodes_[xx];
	table.add( curr );
	cc += curr->symbData_->adjac.getCount();
      }
      cc = cc / nLocalNd_ + 1;
      while( !table.isEmpty() ) { 
	curr = table.removeHead();
	if( curr->symbData_->adjac.getCount() > cc) nodeList.addHead( curr );
	else nodeList.addTail( curr );
      }
    }
    else {
      for( xx = 0 ; xx < nLocalNd_ ; xx++ ) nodeList.addTail( nodes_[xx] ); 
    }

    // MIS
    hit = miss = 0; // global vars set in callback "PromIsAMGDeleteable"
    ierr = mat_loc->CreateMIS(this, nodeList, PromIsAMGDeleteable_new);
    CHKERRQ(ierr); 

    // set children
    for( xx = 0 ; xx < nNodes_ ; xx++ ) {
      curr = nodes_[xx];
      if( curr->getState() == PROM_DELETED ) {
	PromNode *sel = curr->symbData_->parent; 
	ierr = sel->addChild( curr );
      }
      else if(curr->getState() != PROM_SELECTED) {
      }
    }
  }
  else{
#if defined(PROM_HAVE_METIS)
    // get num domains
    int nbglobal = (int)((float)nglobal()/(float)amg_factor_+.5),*pnode_part;
    ierr = PetscMalloc( (nloc+1)*sizeof(int), &pnode_part); CHKERRQ(ierr);
    if(vb>1)PetscPrintf(vb > 2 ? PETSC_COMM_SELF : gComm_,
			"\t[%d]%s %d aggs. amg factor=%d\n",
			comm_.mype(),__FUNCT__,nbglobal,amg_factor_);
    // 'ewgts' causes bug?
    int ii, wt = (ndwts ? 2 : 0) + (ewgts ? 1 : 0);
    int options[5]={0,3,1,1,0},nmf=0,edgecut=-1,ncon=1,npp=comm_.np()/factor_;
    float tpwgts[PROM_BSZ],ubvec=1.2; idxtype vtxdist[PROM_BSZ+1];
    for(ii=0;ii<npp;ii++)tpwgts[ii] = 1./npp;
    for(ii=xx=0;ii<=npp;ii++,xx+=factor_)vtxdist[ii] = proc_gnode_[xx];
    perf_mon_.EventBegin( PromContext::prom_logs_[LOG_METIS] );
    // 
    //ParMETIS_V3_AdaptiveRepart(vtxdist, xadj, adjncy, ndwts, ndwts, ewgts, 
    //		       &wt, &nmf, &ncon, &npp, tpwgts, &ubvec, 
    //		       &itr, options, &edgecut, pnode_part, &gComm_);
    ParMETIS_V3_PartKway( vtxdist, xadj, adjncy, ndwts, ewgts, &wt, &nmf, &ncon,
			  &nbglobal, tpwgts, &ubvec, options, &edgecut, 
			  pnode_part, &gComm_ );
    perf_mon_.EventEnd(PromContext::prom_logs_[LOG_METIS] );
    //
    ierr = PickSubdomains( pnode_part );        CHKERRQ(ierr);
    PetscFree( pnode_part );     
#else
    SETERRQ(1,"Aggregation 'factor' not supported for build w/o METIS");
#endif   
  }

  delete mat_loc; mat_loc = NULL; PetscFree(adjncy); 
  if( ewgts != NULL ) PetscFree( ewgts ); ewgts = NULL; // static
  // make grid row wieghts
  ierr = MakeWeightMatrix( ndwts, &adjncy ); CHKERRQ(ierr);
  PetscFree(ndwts);
  PetscFree(adjncy);
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__    
#define __FUNCT__ "PromAggGrid::ConstructAggs" 
/* PromAggGrid::ConstructAggs ********************************************
 *
 *  - agglomerate aggragates w/ smoothing and make !global ghosts consistant
 *
 *  INPUT:
 *
 *  SIDE EFFECTS:
 *    - constructs selected child lists
 *
 *  RETURN:
 *    - PET cerror code
 *
 */
int PromAggGrid::ConstructAggs()
{
  int ierr,ii,xx,todo=0,tt,nselected,proc,gtag=getNewTag(),mina=0,maxa;
  MPI_Status status; 
  PromTable<int> procTable(157), table1(39), table2(39);
  PromList<MPI_Request*> smessages; 
  PromTable<PromList<PromNode*>*> proc_list(57);
  const int mype=comm_.mype(),my0=my0Nd(),nloc=nlocal(),stag1=getNewTag(),stag2=getNewTag();
  const int vb=options_.verbose_;
  PromList<PromAggNode*> agglist1, agglist2, singletons(1);
  PromTable<int> proc_nsel(57); Position pos;
  PromGlobalID *dep;

  PetscFunctionBegin;  
  // set nchildinsel and minadjinsel
  ierr = SetSelCaches( nselected, global_agg, &proc_nsel); CHKERRQ(ierr);
  for( xx = 0 ; xx < nNodes_ ; xx++ ) {
    PromNode *c2 = nodes_[xx];
    if( c2->isGhost() ) procTable.add( c2->symbData_->ghostID.proc() + 1 ); 
    if( c2->symbData_->dependents ) {
      pos = c2->symbData_->dependents->getHeadPosition();
      while( pos ) { // send to old ghosts
	dep = c2->symbData_->dependents->getNext(pos);
	procTable.add( dep->proc() + 1 );
      }
    } 
  }
  // check for small subdomains -- left in mis 
  bool flag,didit;  double factor; int outl = 1;
  ierr = options_.GetOptionDouble("-aggmg_factor2",&factor,&flag);
  if( !flag || factor < 0.0 ) factor = 1.0;
  for(ii=0;ii<getLevel();ii++) factor /= 1.2;
  const int numerat = (global_agg) ? (int)(factor*(double)ADJAC_FACTOR)+1 : 2;
  for( int kk = 0 ; kk < numerat ; kk++ ){
    if(vb>outl)
      PetscPrintf(gComm_,"[%d]%s %d/%d) nselected=%d. min aggregate sz=%d ",
		  mype,__FUNCT__,kk,numerat,nselected,mina+1);
    // iterate over all SELECTED nodes
    for( xx = maxa = 0, mina = 9999; xx < nloc ; xx++ ) {
      PromNode *curr = nodes_[xx];
      if( curr->getState() == PROM_DELETED ) continue;     // SELECTE'D
      // don't kill: big 1, the last 1 as others think you have 1, done one 
      if( (ii=curr->nChildInSel()) > (kk*curr->minAdjInSel())/ADJAC_FACTOR || 
	  nselected == 1 || curr->isMarked() ){ 
	if( ii > maxa ) maxa = ii;  // good agg
	if( ii < mina ) mina = ii;
      }
      else { 
	assert( nselected > 1);
	// loop thru aggregate, see if other proc can try to dump on me
	PromAggNode *aggn = dynamic_cast<PromAggNode*>(curr); assert(aggn!=NULL);
	bool doit = TRUE;
	for(/*void*/;aggn != NULL && doit; 
		    aggn = dynamic_cast<PromAggNode*>(aggn->next_) ){
	  if( aggn->isGhost() ) { 
	    int nsel = proc_nsel.find(aggn->symbData_->ghostID.proc()+1) - 1;
	    if( nsel < 1 ) doit = FALSE; // break -- do not free ghost w/o place
	  }
	  else { 
	    PromNode *c2 = NULL;
	    PromTable<PromNode*> *adj = &aggn->symbData_->adjac;
	    TablePos tpos = adj->getHeadPosition();  
	    while( tpos ){
	      PromNode *c3 = adj->getNext( tpos );  
	      // matches test, or at least is more conservative
	      if( c3->isGhost() && !global_agg ) continue;
	      if( c3->getNDF() != aggn->getNDF() ) continue;
	      if( c3->getState() == PROM_DELETED ) {
		c3 = c3->symbData_->parent;
		if( c3 == NULL /*|| c3->isGhost()*/ ) continue;
	      }
	      if( c3 == curr ) continue; // no good
	      assert( c3->getState() == PROM_SELECTED );
	      // skip ghosts 
	      if( c3->isGhost() ) {
		// skip new ghost -- dependents->getCount()==0 
		if(c3->symbData_->dependents->getCount() == 0) continue;
		// active ghost
		if( c3->nChildInSel() == 0 ) {
		  int proc = c3->symbData_->ghostID.proc();
		  if( curr->nChildInSel() != 0 || proc > mype ) {
		    doit = FALSE; // break -- it wants me
		    break;
		  }
		}
		else if( curr->nChildInSel()==0 && 
			 c3->nChildInSel() > (kk*c3->minAdjInSel())/ADJAC_FACTOR){
		  /* i want it */;
		}
		else continue; 
	      }
	      if( c3->nChildInSel() > 2*c3->minAdjInSel() && 
		  curr->nChildInSel() > 2 ) continue; // !pile
	      // end match
	      c2 = c3; // flag
	    } // tpos
	    if( c2==NULL ) doit = FALSE; // break -- no good place to put aggn
	  } 
	} // agg list
	// reaggragate
	if( !doit ) {
	  ii = curr->nChildInSel(); 	// no luck
	  if( ii > maxa ) maxa = ii; if( ii < mina ) mina = ii;
	}
	else {
	  // empty aggragate (curr) - get list of locals
	  nselected--; assert( agglist1.isEmpty() ); 
	  while( !curr->isChildListEmpty() ) {
	    PromStack *stc = curr->PopChild();
	    PromNode *c2 = dynamic_cast<PromNode*>(stc); assert(c2!=NULL);
	    c2->symbData_->parent = NULL;
	    // ghosts done w/ update
	    if( !c2->isGhost() ) {
	      PromAggNode *an = dynamic_cast<PromAggNode*>(c2); 
	      agglist1.addTail( an ); 
	    }
	  }
	  const int currNChild = curr->nChildInSel();
	  ierr = curr->SetnChildInSel( 0 );   CHKERRQ(ierr);
	  ierr = curr->SetMinAdjInSel( 0 );   CHKERRQ(ierr);    
	  curr->setState( PROM_DELETED ); 
	  assert( curr->symbData_->parent == NULL );
	  PromAggNode *an = dynamic_cast<PromAggNode*>(curr); assert(an!=NULL);
	  agglist1.addTail( an ); 
	  // empty agglist
	  assert( agglist2.isEmpty() );
	  while( !agglist1.isEmpty() ) {
	    PromAggNode *aggn = agglist1.removeHead();
	    // move aggn
	    ierr = aggn->MigrateToAgg( 0, kk, ewgts, proc_list,
				       didit, nodes_, ghosts_, nloc, my0, 
				       table1, table2, currNChild );
	    CHKERRQ(ierr);
	    // freeze with new
	    if( !didit ) agglist2.addTail( aggn );
	    else if( global_agg ) {
	      PromNode *par = aggn->symbData_->parent; par->mark(); 
	      assert(par->getState() == PROM_SELECTED);	  	    
	    }
	  }
	  while( !agglist2.isEmpty() ){
	    aggn = agglist2.removeHead();
	    // move -- can send back
	    ierr = aggn->MigrateToAgg(0, kk, ewgts, proc_list,
				      didit,nodes_, ghosts_, nloc, my0, 
				      table1, table2, currNChild );
	    CHKERRQ(ierr);
	    if( !didit ) {
	      ierr = aggn->MigrateToAgg( 1, kk, ewgts, proc_list,
					 didit, nodes_, ghosts_, nloc, my0, 
					 table1, table2, currNChild );
	      CHKERRQ(ierr);
	      if( vb > 2 || (!didit && vb > 1) )
		PetscPrintf( (vb > 3 || !didit) ? PETSC_COMM_SELF : gComm_,
			     "\t\t[%d]%s %d) %s hard add gid=%d nc=%d\n",
	 		     mype, __FUNCT__, kk, 
			     didit ? "Warning" : "WARNING VERY ",
			     aggn->index_.global(), currNChild );
	      if( !didit ){
		for(int xx = 0 ; xx < nloc ; xx++ ){
		  PromNode *c2 = nodes_[xx];
		  if(c2->getState() == PROM_SELECTED && c2->getNDF()==aggn->getNDF()) {
		    ierr = aggn->MigrateTo(c2,proc_list,table2);CHKERRQ(ierr);
		    didit = TRUE;
		    break;
		  }
		}
	      }
	    }
	    if(!didit){SETERRQ1(1,"reaggregation error gid=%d",aggn->index_.global());}
	    else if( global_agg ) {
	      PromNode *par = aggn->symbData_->parent; par->mark(); 
	      assert(par->getState() == PROM_SELECTED);	  	    
	    }
	  }
	}
      }
    } // node loop
    if(vb>outl)PetscPrintf(gComm_,"*");
    // resolve data structures
    if( !global_agg ) {
      if(vb>outl)PetscPrintf(gComm_,"*\n");
      assert( proc_list.isEmpty() );
    }
    else { // send updates send(stag1)
      ierr = SendAggUpdates( proc_list, smessages, stag1, kk ); CHKERRQ(ierr);
      assert(todo==0); assert( agglist1.isEmpty() );
      if(vb>outl)PetscPrintf(gComm_,"*");
      // collect messages and fix ghosts (can strand local DELETEDs, todo)
      ierr = CapAndRecvUpdates( procTable, stag1, gtag, todo, agglist1, kk );
      CHKERRQ(ierr);
      if(vb>outl)PetscPrintf(gComm_,"*");
      // cleanup sends
      while( !smessages.isEmpty() ) {
	MPI_Request *sreq = smessages.removeHead();
	MPI_Test( sreq, &ii, &status );
	if( ii ) { MPI_Wait( sreq, &status );    PetscFree( sreq ); }
	else smessages.addTail( sreq );
      }
      // 1) fix local lost from ghost updates
      while( !agglist1.isEmpty() ) {
	PromAggNode *aggn = agglist1.removeHead();
	if( aggn->isGhost() ) continue; // done by others
	if( aggn->symbData_->parent == NULL ) {
	  // move -- can send back
	  ierr = aggn->MigrateToAgg(1, kk, ewgts, proc_list, didit, 
				    nodes_, ghosts_, nloc, my0, table1,table2);
	  CHKERRQ(ierr);
	  if( !didit ) agglist2.addTail( aggn );
	} 
	else assert( aggn->symbData_->parent->getState() == PROM_SELECTED );
      }
      // 2) 
      while( !agglist2.isEmpty() ){
	PromAggNode *aggn = agglist2.removeHead();
	// move -- can send back
	ierr = aggn->MigrateToAgg(1, kk, ewgts, proc_list, didit, 
				  nodes_, ghosts_, nloc, my0, table1, table2 );
	CHKERRQ(ierr);
	if( !didit ) agglist1.addTail(aggn);
      }
      // 3) 
      while( !agglist1.isEmpty() ){
	PromAggNode *aggn = agglist1.removeHead();
	// move -- can send back
	ierr = aggn->MigrateToAgg( 1, kk, ewgts, proc_list, didit,
				   nodes_, ghosts_, nloc, my0, table1, table2);
	CHKERRQ(ierr);
	if( vb > 1 ) {
	  PetscPrintf((vb > 2) ? PETSC_COMM_SELF : gComm_,
		      "\t[%d]%s %d) %s: hard add of recv. nsel=%d\n",
		      mype,__FUNCT__,aggn->index_.global(), didit ? " ":"Warning",nselected);
	}
	if( !didit ) {
	  PromTable<PromNode*> *adj = &aggn->symbData_->adjac;
	  TablePos tpos = adj->getHeadPosition();  
	  while( tpos && !didit ){ 
	    PromNode *c4 = adj->getNext(tpos), *c3 = c4->symbData_->parent;  
	    if(c4->getState()==PROM_DELETED && c3!=aggn && c3!=NULL &&
	       !c3->isGhost() ) c4 = c3;
	    if(c4->getState()==PROM_SELECTED && !c4->isGhost() && aggn->getNDF()==c4->getNDF()){
	      // doit
	      ierr = aggn->MigrateTo( c4, proc_list, table2 ); CHKERRQ(ierr);
	      didit = TRUE;
	      break;
	    }
	    else {
	      PromTable<PromNode*> *adj2 = &c4->symbData_->adjac;
	      TablePos tpos2 = adj2->getHeadPosition();  
	      while( tpos2 ){
		PromNode *c2 = adj2->getNext( tpos2 );
		if(c2 == aggn) continue;
		if(c2->getState()==PROM_DELETED) c2 = c2->symbData_->parent;
		if(c2==NULL||c2->isGhost()||c2==aggn||adj->find(c2)) continue;
		if( aggn->getNDF() != c2->getNDF() ) continue;
		// doit
		ierr = aggn->MigrateTo( c2, proc_list, table2 ); CHKERRQ(ierr);
		didit = TRUE;
		break;
	      }
	    }
	  }
	  if( !didit ){
	    if( vb > 0 ) {
	      PetscPrintf(PETSC_COMM_SELF,
"\t[%d]%s ** WARNING ***: create singleton gid=%d it=%d/%d ndf=%d\n",
			  mype,__FUNCT__,aggn->index_.global(),
			  kk,numerat,aggn->getNDF());
	    }
	    singletons.addTail( aggn ); didit = TRUE;
            assert(aggn->symbData_->parent==NULL);
	  }
        }
      }
      if(vb>outl)PetscPrintf(gComm_,"*");
      // check singletons
      pos = singletons.getHeadPosition();
      while( pos ) { // send to old ghosts
	PromAggNode *aggn = singletons.getNext( pos );
	TablePos tpos = aggn->symbData_->adjac.getHeadPosition(); 
	while( tpos && pos ){
	  PromNode *c2 = aggn->symbData_->adjac.getNext(tpos);
          PromAggNode *cc = dynamic_cast<PromAggNode*>(c2);   assert(c2!=aggn);
	  Position pos2;
	  if((pos2=singletons.find(cc))==NULL && c2->getState()==PROM_DELETED){
	    c2 = c2->symbData_->parent;
          }
	  if(c2 != NULL && c2 != aggn && !c2->isGhost() && c2->getNDF() == aggn->getNDF()){
	    Position pos3 = singletons.find( aggn );
	    if( pos3 == NULL ) {
	      if(vb>0)PetscPrintf(PETSC_COMM_SELF,
				  "\t\t\t[%d]%s pos2==NULL ??? count=%d %d\n",
				  mype,__FUNCT__,singletons.getCount(),
				  aggn->index_.global());
	    }
	    else{
	      if(!isBase())continue; // couple on fine grid only ???
	      if(vb>0)PetscPrintf(PETSC_COMM_SELF,
				  "\t[%d]%s fix sing. (it=%d) count=%d %d\n",
				  mype,__FUNCT__,kk,singletons.getCount(),
				  aggn->index_.global());
	      singletons.removeAt( pos3 );
	      // 
	      if( pos2 != NULL ) {
		singletons.removeAt( pos2 );
		c2->setState( PROM_SELECTED );
		if(vb>0)PetscPrintf(PETSC_COMM_SELF,"\t[%d]%s couple %d\n",
				    mype,__FUNCT__,aggn->index_.global());
	      }
	      else assert(c2->getState()==PROM_SELECTED);
	      assert(c2->symbData_->parent==NULL);
	      // sends messages 
	      assert(aggn->getState()==PROM_DELETED); 
	      ierr = aggn->MigrateTo( c2, proc_list, table2 ); CHKERRQ(ierr);
	      //
	      while( (pos == pos2 || pos == pos3) && pos != NULL ) {
		if(vb>0)PetscPrintf(PETSC_COMM_SELF,"\t[%d]%s skip %d\n",
				    mype,__FUNCT__,aggn->index_.global());
		singletons.getNext( pos ); // this could be next,so skip
	      }
	      break;
	    }
	  }
	}
	if( singletons.isEmpty() && pos != NULL ) {
	  PetscPrintf(PETSC_COMM_SELF,
		      "\t\t[%d]%s ERROR: singletons.isEmpty() && pos!=NULL\n",
		      mype,__FUNCT__);
	}
      }
      // send updates back
      ierr = SendAggUpdates( proc_list, smessages, stag2, kk ); CHKERRQ(ierr);
      if(vb>outl)PetscPrintf(gComm_,"*");
      ierr = CapAndRecvUpdates( procTable, stag2, gtag, ii, agglist1, kk ); 
      CHKERRQ(ierr);
      todo += ii;  assert(proc_list.isEmpty()); assert(agglist1.isEmpty());
      // cleanup sends
      while( !smessages.isEmpty() ) {
	MPI_Request *sreq = smessages.removeHead();
	MPI_Test( sreq, &ii, &status );
	if( ii ) { MPI_Wait( sreq, &status );    PetscFree( sreq ); }
	else smessages.addTail( sreq );
      }
      // collect exta ghosts -- unmark them
      if(vb>outl)PetscPrintf(gComm_,"*");
      ierr = ReceiveAddGhost( todo, gtag, procTable ); CHKERRQ(ierr);
      todo = 0; gtag = getNewTag();
      // collect exta ghosts and deps into 'procTable'
      ierr = AddDepsToList( procTable ); CHKERRQ(ierr);
      // unmark -- new ghosts
      for(xx=0;xx<nNodes_;xx++) nodes_[xx]->mark( FALSE );
      if(vb>outl)PetscPrintf(gComm_,"*\n");
    } // global_agg
  } // for agg factor..., iterator for killing small aggs
  if( nselected > 0 && (vb>2 || (vb>0 && mina < 1) ) ) {
    PetscPrintf( mina <= 1 ? PETSC_COMM_SELF : PETSC_COMM_WORLD,
		 "[%d]%s Agg. size min=%d max=%d. nselected=%d\n",
		 mype,__FUNCT__,mina+1,maxa+1,nselected);
  }
  assert( proc_list.isEmpty() ); assert(todo == 0);
  if(vb>1)PetscPrintf( gComm_,"[%d]%s nselected=%d Agg. size min=%d\n",
		       mype,__FUNCT__,nselected,mina+1);

  // set singletons selected
  while( !singletons.isEmpty() ) {
    PromAggNode *aggn = singletons.removeHead();
    aggn->setState( PROM_SELECTED );
    nselected++; 
  }

  // clean up
  if( global_agg ) {
    // see if ghosts are not set (singletons)
    for( xx = nloc ; xx < nNodes_ ; xx++ ){
      PromNode *c2 = nodes_[xx];
      if( c2->symbData_->parent==NULL && c2->getState()==PROM_DELETED) {
	c2->setState( PROM_SELECTED );
      }
    }
  }
  else {
    // globalize
    PromList<PromNode*> *list;  
    const int reqsz = sizeof(MPI_Request)/sizeof(int);
    // globalize
    for( xx = 0 ; xx < nloc ; xx++ ){
      PromNode *curr = nodes_[xx]; 
      if( curr->isBound() ){
	pos = curr->symbData_->dependents->getHeadPosition();  
	while ( pos ) {
	  dep = curr->symbData_->dependents->getNext( pos ); 
	  proc = dep->proc(); 
	  if( (list=proc_list.find(proc+1)) == NULL ) {
	    list = new PromList<PromNode*>(100); CHKNEW(list,1);
	    proc_list.add( list, proc+1 );
	  }
	  list->addTail( curr ); 
	}
      }
    }
    int tag = getNewTag();
    // send 
    {
      int *buff, *pb, np = 0, gid, gidS, parts[PROM_BSZ];
      while( !proc_list.isEmpty() ) {
	list = proc_list.removeHead( &proc ); proc--;
	parts[np++] = proc/factor_;
	tt = 2*list->getCount(); MPI_Request *sreq;
	ierr = PetscMalloc((tt+reqsz)*sizeof(int), &sreq); CHKERRQ(ierr); 
	smessages.addTail( sreq );
	buff = (int*)sreq + reqsz; pb = buff;
	while( !list->isEmpty() ) {
	  PromNode *curr = list->removeHead(); 
	  *pb++ = curr->index_.global(); 
	  if( curr->getState() != PROM_DELETED ) *pb++ = -1;
	  else *pb++ = curr->symbData_->parent->index_.global();
	}
	assert(tt == pb-buff);
	// send(tag) *< gid, Sgid/-1 >
	MPI_Isend( buff, tt, MPI_INT, proc/factor_, tag, gComm_, sreq );
	delete list; 
      }
      // recv
      for( xx = nloc, ii = 0 ; xx < nNodes_ ; xx++ ) {
	if( nodes_[xx]->isDead() ) continue; // new ghosts are marked
	ii++;
      } 
      assert(todo==0);
      qsort( (void*)parts, np, sizeof(int), prom_intcompar );
      for( xx = 0 ; xx < np ; xx++ ) {
	MPI_Probe( parts[xx], tag, gComm_, &status ); // det!!!
	MPI_Get_count( &status, MPI_INT, &tt ); 
	proc = parts[xx] * factor_; // local on other proc
	ierr = PetscMalloc(tt*sizeof(int), &buff); CHKERRQ(ierr);
	pb = buff; 
	// receive(tag) *< gid, Sgid/-1 >
	MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag, gComm_, &status);
	ii -= tt/2; assert(ii>=0);
	while( pb-buff < tt ) {
	  gid = *pb++; gidS = *pb++; 
	  PromNode *curr = ghosts_.find(gid+1); assert(curr!=NULL);
	  if( gidS >= 0 ) {
	    curr->setState( PROM_DELETED );
	    PromNode *c2; 
	    todo += addGhost( c2, proc, gidS, gidS - proc_gnode_[proc], gtag );
	    curr->symbData_->parent = c2; c2->addChild( curr );
	  }
	  else{
	    curr->setState(PROM_SELECTED);
	    assert(curr->symbData_->parent==NULL);
	  } 
	}
	assert(pb-buff == tt);	PetscFree(buff);
      }
      assert( ii == 0 );
    }
    // cleanup sends
    while( !smessages.isEmpty() ) {
      MPI_Request *sreq = smessages.removeHead();
      MPI_Wait( sreq, &status );    PetscFree( sreq );
    }
    MPI_Iprobe( MPI_ANY_SOURCE, tag, gComm_, &ii, &status );assert(!ii);
    // collect exta ghosts -- unmark them
    if(vb>1)PetscPrintf(gComm_,"[%d]%s %d new gs\n",mype,__FUNCT__,todo);
    ierr = ReceiveAddGhost( todo, gtag, procTable ); CHKERRQ(ierr);
    todo = 0; gtag = getNewTag(); 
    // collect exta ghosts and deps into 'procTable'
    ierr = AddDepsToList( procTable ); CHKERRQ(ierr);
  }

  // clear 
  for( ii = xx = 0 ; xx < nNodes_ ; xx++ ) {
    PromNode *curr = nodes_[xx];
    assert(!curr->isDead()); assert(!curr->isMarked()); 
    if(curr->getState()==PROM_SELECTED && curr->nChildInSel()==0 && xx<nloc){
      PromTable<PromNode*> *cadj = &curr->symbData_->adjac;
      if( options_.verbose_ > 1 && cadj->getCount() > 0 ) {
	TablePos tpos = cadj->getHeadPosition(); assert(tpos!=0);int nn=0,tt=0;
	while( tpos ){
	  PromNode *c3 = cadj->getNext(tpos);  
	  if( c3->getState() == PROM_DELETED ) {
	    c3 = c3->symbData_->parent; assert(c3 != NULL);
	    nn += c3->nChildInSel();
	  }
	  if( !c3->isGhost() ) tt++;
	} 
	PetscPrintf(PETSC_COMM_SELF,
		    "\t[%d] %s agg size = 1 (gid=%d), %d adjac, %d local adjacs, w/ %d child\n",
		    mype,isBase()?"** WARNING **":"Warning", 
		    my0+xx,curr->symbData_->adjac.getCount(),tt,nn);
      }
      if(cadj->getCount()==0)ii++;
    }
    ierr = curr->SetnChildInSel(0);   CHKERRQ(ierr);  
    ierr = curr->SetMinAdjInSel(0);  CHKERRQ(ierr);  
  }
  if( ii > 0 && options_.verbose_ > 1 ) {
    PetscPrintf(PETSC_COMM_SELF,"\t[%d]%s have %d zero neighbor nodes\n",
		mype,__FUNCT__,ii);
  }
  // debug
  MPI_Iprobe( MPI_ANY_SOURCE, stag1, gComm_, &ii, &status );assert(!ii);
  MPI_Iprobe( MPI_ANY_SOURCE, stag2, gComm_, &ii, &status );assert(!ii);
  MPI_Iprobe( MPI_ANY_SOURCE, gtag, gComm_, &ii, &status );assert(!ii);

  PetscFunctionReturn(0);
}
 
#undef __FUNCT__    
#define __FUNCT__ "PromAggGrid::CapAndRecvUpdates" 
/* PromAggGrid::CapAndRecvUpdates *******************************************
 *
 *  - recv. <NP, gid[S|D]->D(ghost), gidS(loc/gh), dep_proc[NP] >
 *
 *  INPUT: 
 *    - procTable: (in/out)
 *    - tag
 *    - gtag
 *    - todo: (out)
 *    - lostlist: (out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code
 */
int PromAggGrid::CapAndRecvUpdates( PromTable<int> &procTable, const int tag, 
				    const int gtag, int &todo, 
				    PromList<PromAggNode*> &lostlist, 
				    const int kk )
{
  int ii,tt,ierr,dgid,np,sgid,ndf,proc2,*pb,*buff,ptodo;
  PromNode *c1, *c2;
  MPI_Status status; 
  const int my0 = proc_gnode_[comm_.mype()]; assert( lostlist.isEmpty() );

  todo = 0; // out
  // capper(tag)
  TablePos tipos = procTable.getHeadPosition(); 
  while( tipos ){ 
    int proc = procTable.getNext(tipos) - 1;
    MPI_Send( NULL, 0, MPI_INT, proc/factor_, tag, gComm_ );
  }

  // receive fixes
  ptodo = procTable.getCount();
  while( ptodo ) {
    MPI_Probe( MPI_ANY_SOURCE, tag, gComm_, &status ); // det!!!
    MPI_Get_count( &status, MPI_INT, &tt ); 
#if defined(PETSC_USE_BOPT_g)
    assert(procTable.findi(status.MPI_SOURCE*factor_+1));
#endif
    if( tt == 0 ){
      MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE, tag, gComm_, &status);
      ptodo--;
    }
    else {
      ierr = PetscMalloc(tt*sizeof(int), &buff ); CHKERRQ(ierr);
      // recv(tag) *< NP, gid[S|D]->D(ghost), gidS(loc/gh), ndf, *dep_proc >, kk
      MPI_Recv(buff, tt, MPI_INT, status.MPI_SOURCE, tag, gComm_, &status);
      pb = buff; 
      while( pb-buff < tt-1 ) {
	np = *pb++; dgid = *pb++; sgid = *pb++; ndf = *pb++;
	// read procs
	for(int k=0;k<np; k++){
	  int pp = *pb++;
	  assert(pp < comm_.np() && pp >= 0);
	  if( pp != comm_.mype() ) procTable.add(pp+1); // symmetric
	}
	// c1(ghost): [S|D] -> D
	ierr = FindProc(proc_gnode_,comm_.np(),dgid,proc2); CHKERRQ(ierr);
	ii = dgid - proc_gnode_[proc2];	assert(proc2!=comm_.mype());
	int add = addGhost( c1, proc2, dgid, ii, gtag ); todo += add;
	if( add ) c1->setState( PROM_DELETED ); 
	if( add ) c1->SetNDF( ndf );
	else assert( c1->getNDF() == ndf );
	// get SELECTED "c2"
	ii = sgid - my0;
	if( ii >= 0 && ii < nLocalNd_ ) c2 = nodes_[ii];
	else{ // secondary recv
	  ierr = FindProc(proc_gnode_,comm_.np(),sgid,proc2); CHKERRQ(ierr);
	  int ii = sgid - proc_gnode_[proc2]; 	assert(proc2!=comm_.mype());
	  int add2 = addGhost( c2, proc2, sgid, ii, gtag ); todo += add2;
	  if( add2 ) c2->setState( PROM_SELECTED );
	  if( add2 ) c2->SetNDF( ndf );
	  else assert( c2->getNDF() == ndf );
	}
	if( c2->getState() != PROM_SELECTED ) {
	  //int proc = status.MPI_SOURCE * factor_;
	  //PetscPrintf(PETSC_COMM_SELF,"\t[%d]%s Warning gid %d from %d, %s node ('coupled' fix)\n",comm_.mype(),__FUNCT__,c2->index_.global(), proc,(ii >= 0 && ii < nLocalNd_) ? "local" : "ghost secondary recv");
	  /* debug
	     int b[2] = {sgid,dgid};
	     MPI_Send( b, 2, MPI_INT, status.MPI_SOURCE, 99999, gComm_ ); 
	     PetscSleep(2);
	     SETERRQ2(1,"c2->getState() != PROM_SELECTED, %s recv. sgid=%d",
	     (ii>=0 && ii<nLocalNd_) ? "primary" :"secondary",sgid);*/
          c2->setState(PROM_SELECTED);
       	}
	// delete selected node
	if( c1->getState() == PROM_SELECTED ){
	  // delete selected ghost
	  while( !c1->isChildListEmpty() ){
	    assert( c1->getState() == PROM_SELECTED ); // 
	    PromStack *stc = c1->PopChild(); assert(stc != NULL);
	    PromAggNode *c3 = dynamic_cast<PromAggNode*>(stc);assert(c3!=NULL);
	    lostlist.addTail( c3 );	  c3->symbData_->parent = NULL;
	  }
	  // SELECTED --> DELETED
	  ierr = c1->SetnChildInSel(0);   CHKERRQ(ierr);  
	  ierr = c1->SetMinAdjInSel(0);  CHKERRQ(ierr);
	  c1->setState( PROM_DELETED ); 
	}
	else if( c1->symbData_->parent != NULL ) {
	  // remove deleted
	  ii = c1->symbData_->parent->removeChild( c1 );   assert(ii);
	  c1->symbData_->parent = NULL;
	}
	c1->symbData_->parent = c2; c2->addChild( c1 );
	assert(c2->getNDF() == c1->getNDF());
	ierr = c2->SetnChildInSel( c2->nChildInSel() + 1 );  CHKERRQ(ierr);  
      }
      int rkk = *pb++;
      assert(pb-buff == tt);      PetscFree(buff);
      assert(kk==rkk);
    }
  }
  MPI_Iprobe( MPI_ANY_SOURCE, tag, gComm_, &ii, &status );     assert(!ii);

  /* debug 
  PetscSleep(1);
  MPI_Iprobe( MPI_ANY_SOURCE, 99999, gComm_, &ii, &status );     
  if(ii){
    int b[2],sgid,dgid,slid,dlid;
    MPI_Recv(b, 2, MPI_INT, status.MPI_SOURCE, 99999, gComm_, &status);
    sgid=b[0]; dgid=b[1]; slid=sgid-my0; dlid=dgid-my0;
    PromNode *s1,*d2;
    if(slid>=0 && slid<nLocalNd_) s1 = nodes_[slid];
    else s1 = ghosts_.find(sgid+1);
    assert(s1!=NULL);
    if(dlid>=0 && dlid<nLocalNd_) d2 = nodes_[dlid];
    else d2 = ghosts_.find(dgid+1);
    assert(d2!=NULL);
    PetscPrintf(PETSC_COMM_SELF,
		"\n\t\t[%d] recv sgid %d(state=%s), dgid=%d(state=%s), from %d\n",
		comm_.mype(),
		sgid, s1->getState()==PROM_SELECTED?"selected":"deleted",
		dgid, d2->getState()==PROM_SELECTED?"selected":"deleted",
		status.MPI_SOURCE);
    if( s1->getState()==PROM_DELETED ) {
      PromNode *s2 = s1->symbData_->parent; assert(s2 != NULL);
      PetscPrintf(PETSC_COMM_SELF,
		  "\t\t\t[%d] deleted sgid's parent is %d (isgh=%d)\n",
		  comm_.mype(),s2->index_.global(),
		  s2->isGhost() ? c2->symbData_->ghostID.proc() : -1);
    }
  }
  MPI_Barrier(gComm_); */

  return 0;
} 

#undef __FUNCT__    
#define __FUNCT__ "PromAggGrid::SendAggUpdates" 
/* PromAggGrid::SendAggUpdates *********************************************
 *
 *  - 
 *
 *  INPUT: 
 *    - proc_list: (in/out)
 *    - smessages: (in/out)
 *    - tag
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code
 */
int PromAggGrid::SendAggUpdates( PromTable<PromList<PromNode*>*> &proc_list, 
				 PromList<MPI_Request*> &smessages, 
				 const int tag, const int kk  )
{
  int ierr,ii,tt,proc;
  int  reqsz = sizeof(MPI_Request)/sizeof(int);
  PromList<PromNode*> *list;
  PromNode *del, *sel;
  PromTable<int> table(57); 
  PromGlobalID *dep;

  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--;
    // count size
    tt = 1;
    Position pos = list->getHeadPosition(); 
    while( pos ) {
      del = list->getNext( pos );      
      sel = list->getNext( pos ); 
      assert( table.isEmpty() );
      if( sel->isExtended() ) { 
	Position pos = sel->symbData_->dependents->getHeadPosition();
	assert(pos!=NULL);
	while( pos ) {
	  dep = sel->symbData_->dependents->getNext( pos );
	  ii = table.add( dep->proc()+1 ); tt++; assert(ii);
	}
      }
      if( del->isExtended() ) {
	Position pos = del->symbData_->dependents->getHeadPosition();  
	assert(pos!=NULL);
	while ( pos ) {
	  dep = del->symbData_->dependents->getNext( pos );
	  if( table.add( dep->proc()+1 ) ) tt++;
	}
      }
      assert( !table.isEmpty() ); 
      while( !table.isEmpty() ) table.removeHead();
    }

    tt += 4*(list->getCount()/2); assert(list->getCount()%2==0);// gids(3) + NP
    if( tt == 0 ) continue; 
    assert( list->getCount() );
    MPI_Request *sreq;  
    ierr = PetscMalloc( (tt+reqsz)*sizeof(int), &sreq ); CHKERRQ(ierr);
    smessages.addTail( sreq );     
    int *buff = (int*)sreq + reqsz, *pb = buff, *pb2;
    while( !list->isEmpty() ) {
      del = list->removeHead();
      sel = list->removeHead(); assert(sel==del->symbData_->parent);
      pb2 = pb; *pb++ = -1; // save for NP
      *pb++ = del->index_.global(); *pb++ = sel->index_.global();
      *pb++ = del->getNDF();  assert(sel->getNDF() == del->getNDF());
      // write procs and make proc list
      assert( table.isEmpty() );
      if( sel->isExtended() ) {
	Position pos = sel->symbData_->dependents->getHeadPosition();
	while ( pos ) {
	  dep = sel->symbData_->dependents->getNext( pos );
	  table.add( dep->proc()+1 ); *pb++ = dep->proc(); 
	}
      }
      if( del->isExtended() ) {
	Position pos = del->symbData_->dependents->getHeadPosition();
	while ( pos ) {
	  dep = del->symbData_->dependents->getNext( pos );
	  if( table.add( dep->proc()+1 ) ) *pb++ = dep->proc();
	}
      }
      *pb2 = table.getCount(); 
      while( !table.isEmpty() ) table.removeHead();
    }
    *pb++ = kk;
    assert(tt == pb-buff);
    // send(tag) *< NP, gid[S|D]->D(loc/gho), gidS(loc/gho), ndf, *dep_proc >, kk
    MPI_Isend( buff, tt, MPI_INT, proc/factor_, tag, gComm_, sreq );
    delete list; 
  }
  
  return 0;
}
 
#undef __FUNCT__    
#define __FUNCT__ "PromAggNode::MigrateToAgg" 
/* PromAggNode::MigrateToAgg *********************************************
 *
 *  - move this (local deleted) to other domain
 *
 *  INPUT: 
 *    - hard:
 *    - kk:
 *    - ewgts:
 *    - proc_list: add messages <selected,deleted> (in/out)
 *    - didit: (out)
 *    - nodes: (in/out)
 *    - ghosts:
 *    - nlocal:
 *    - my0:
 *    - node_wt: (buffer)
 *    - table: (buffer)
 *    - nchildInSel:
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code
 */
int PromAggNode::MigrateToAgg( const unsigned int hard, const int kk,
			       const int * const ewgts, 
			       PromTable<PromList<PromNode*>*> &proc_list,
			       bool &didit, 
			       PromNode *nodes[],
			       const PromTable<PromNode*> &ghosts, 
			       const int nlocal,
			       const int my0, 
			       PromTable<int> &node_wt, 
			       PromTable<int> &table, 
			       const int nChildInCur /*=1000*/)
{
  assert(getState()==PROM_DELETED && !isGhost());
  int ierr, ii, tt, wmax=0, wht; 
  const int *whts = ewgts ? &ewgts[xadj[index_()]] : NULL;

  // get best agg 
  PromNode *c2 = NULL;
  PromTable<PromNode*> *adj = &symbData_->adjac;
  TablePos tpos = adj->getHeadPosition(); assert(tpos!=0);
  while( tpos ){
    PromNode *c3 = adj->getNext(tpos); 
    // matches test!!! (when hard<1 && nChildInCur != 1000)
    if( c3->getNDF() != getNDF() ) continue;
    if( c3->isGhost() && !global_agg ) continue;
    wht = (whts!=NULL) ? *whts++ : 10; // matches 'SetAMGEdgeWeights'
    if( c3->getState() == PROM_DELETED ) {
      c3 = c3->symbData_->parent; // get sel
      if( c3 == NULL /*|| c3->isGhost()*/ ) continue;// no good, not curr
    }
    assert( c3->getState() == PROM_SELECTED );
    if( c3->isGhost() ) {
      // skip new ghost -- dependents->getCount()==0 
      if( c3->symbData_->dependents->getCount()==0 ) continue;
      // active ghost
      if( nChildInCur != 0 ) continue;    // skip ghosts 
      else if( c3->nChildInSel() > (kk*c3->minAdjInSel())/ADJAC_FACTOR ){
	/* i want it */;
      }
      else continue; // i am zero, but other is active
    }
    if(hard==0 && c3->nChildInSel() > 2*c3->minAdjInSel() && nChildInCur > 2){
      continue; // !pile
    }
    // end match
    // have one, add weight
    ii = c3->index_.global() + 1;  tt = node_wt.find( ii ); // old wt. sum
    node_wt.add( tt + wht + 1, ii ); /* can overwrite */ 
    c2 = c3; // flag
  }
  assert(whts==NULL || xadj[index_()+1]==whts-ewgts);
  
  // get best (c2) max sum == wmax
  if( !node_wt.isEmpty() ){
    assert(c2 != NULL); c2 = NULL; PromNode *c3;
    do {
      int gid, wht = node_wt.removeHead( &gid ) - 1; gid--; 
      int lid = gid - my0;
      if( lid >= 0 && lid < nlocal ) c3 = nodes[lid];
      else c3 = ghosts.find(gid+1); 
      if( (c2 == NULL || wht > wmax) && c3 != NULL ) { c2 = c3; wmax = wht;  }
    }while( !node_wt.isEmpty() );
  }

  // move it
  if( c2 == NULL ) {
    didit = FALSE; // out 
  }
  else{
    didit = TRUE; // out
    ierr = MigrateTo( c2, proc_list, table );   CHKERRQ(ierr);
  }
  assert( node_wt.isEmpty() && table.isEmpty());
  
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromAggNode::MigrateTo" 
/* PromAggNode::MigrateTo *********************************************
 *
 *  - move this (local deleted) to other domain (c2)
 *
 *  INPUT: 
 *    - c2:
 *    - proc_list: add messages <selected,deleted> (in/out)
 *    - table: (buffer)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code
 */
int PromAggNode::MigrateTo( PromNode *c2, 
			    PromTable<PromList<PromNode*>*> &proc_list,
			    PromTable<int> &table )
{
  int ierr,ii,proc;
  PromList<PromNode*> *list; 
  Position pos;

  if( symbData_->parent ){
    int rem = symbData_->parent->removeChild(this); assert(rem);
    symbData_->parent = NULL;
  }
  c2->addChild( this ); symbData_->parent = c2;
  if(c2->getNDF() != getNDF()) SETERRQ2(1,"c2->ndf(%d) != ndf(%d)",c2->getNDF(),getNDF());
  ierr = c2->SetnChildInSel( c2->nChildInSel() + 1 );  CHKERRQ(ierr);

  // c2/c1 is bound/ghost updates
  if( (c2->isExtended() || isExtended()) && global_agg ) {
    assert( table.isEmpty() );
    // make list of sendies
    if( c2->isExtended() ) { 
      pos = c2->symbData_->dependents->getHeadPosition();  assert(pos!=NULL);
      while ( pos ) {
	PromGlobalID *dep = c2->symbData_->dependents->getNext( pos );
	ii = table.add( dep->proc()+1 ); assert(ii);
      }
    }
    if( isExtended() ) {
      pos = symbData_->dependents->getHeadPosition();  assert(pos!=NULL);
      while ( pos ) {
	PromGlobalID *dep = symbData_->dependents->getNext( pos );
	table.add( dep->proc()+1 );
      }
    }
    // send 
    while( !table.isEmpty() ) {
      proc = table.removeHead() - 1;       
      if( (list=proc_list.find(proc+1)) == NULL ) {
	list = new PromList<PromNode*>(100); CHKNEW(list,1);
	proc_list.add( list, proc+1 );
      }  
      list->addTail( this ); list->addTail( c2 );
    }
  }
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromAggGrid::SetSelCaches" 
/* PromAggGrid::SetSelCaches ********************************************
 *
 *  INPUT:
 *    - nselected: 
 *    - global_agg  
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PET cerror code
 *
 */
int PromAggGrid::SetSelCaches( int &nselected, 
			       const bool global_agg, 
			       PromTable<int> *proc_nsel )
{
  int           ii,xx,tt,todo,proc,ierr; const int tag = getNewTag(); 
  PromNode     *curr,*c1;
  PromList<PromNode*> *list; 
  PromList<MPI_Request*> smessages; 
  PromTable<PromList<PromNode*>*>   proc_list(57);
  const int     reqsz = sizeof(MPI_Request)/sizeof(int);
  MPI_Status    status;
  PromGlobalID *dep;

  // set nchild
  for( xx = nselected = 0 ; xx < nNodes_ ; xx++ ) {
    curr = nodes_[xx];
    curr->kill(FALSE);
    if( curr->getState() == PROM_SELECTED ){
      assert(curr->minAdjInSel()==0);assert(curr->nChildInSel()==0);// no overw
    }
    if( !curr->isGhost() && curr->getState() == PROM_SELECTED ){ // init caches
      nselected++;
      ierr = curr->SetnChildInSel( curr->numChildren() );  CHKERRQ(ierr);  
      c1 = curr;
      while( 1 ){
	if( !c1->isGhost() ) {
	  ii = c1->symbData_->adjac.getCount();
	  if( curr->minAdjInSel() == 0 || ii < curr->minAdjInSel() ) {
	    ierr = curr->SetMinAdjInSel( ii );  CHKERRQ(ierr);
	  }
	}
	PromStack *stc = c1->getChild(); 
	if( stc == NULL ) break;
	c1 = dynamic_cast<PromNode*>(stc); assert(c1!=NULL);
      }
      if( curr->symbData_->dependents && global_agg ) {
	Position pos = curr->symbData_->dependents->getHeadPosition(); 
	while( pos ) { // send to old ghosts
	  dep = curr->symbData_->dependents->getNext(pos);   
	  proc = dep->proc(); 
	  if( (list=proc_list.find(proc+1)) == NULL ) {
	    list = new PromList<PromNode*>(100); CHKNEW(list,1);
	    proc_list.add( list, proc+1 );
	  }
	  list->addTail( curr ); 
	  list->addTail( (PromNode*)dep ); // hack
	}
      }
    }
  }

  // send/recv
  if( global_agg ) {
    // send selected data
    while( !proc_list.isEmpty() ) {
      list = proc_list.removeHead( &ii ); ii--;
      const int proc = ii;
      tt = 3*(list->getCount()/2) + 1; assert(list->getCount()%2 == 0);
      MPI_Request *sreq; ierr = PetscMalloc( (tt+reqsz)*sizeof(int), &sreq ); 
      smessages.addTail( sreq );     CHKERRQ(ierr);
      int *buff = (int*)sreq + reqsz, *pb = buff;
      *pb++ = nselected;
      while( !list->isEmpty() ) {
	curr = list->removeHead();
	dep = (PromGlobalID*)list->removeHead(); // hack
	*pb++ = dep->local(); *pb++ = curr->minAdjInSel();
	*pb++ = curr->nChildInSel();
      } 
      assert(tt == pb-buff);
      // send(tag) nsel, *< lid, minadj, nchild >
      MPI_Isend( buff, tt, MPI_INT, proc/factor_, tag, gComm_, sreq );
      delete list;
    }
    // count recv
    for( xx = nLocalNd_, todo = 0 ; xx < nNodes_ ; xx++ ) {
      curr = nodes_[xx];
      if( curr->getState() != PROM_SELECTED ) continue;
      todo++;
    }
    // recv
    while( todo ) {
      MPI_Probe( MPI_ANY_SOURCE, tag, gComm_, &status );
      MPI_Get_count( &status, MPI_INT, &tt ); 
      int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb = buff;
      CHKERRQ(ierr);
      // recv(tag) nsel, *< lid, minadj, nchild >
      MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag, gComm_, &status );
      int nsel = *pb++;
      int add = proc_nsel->add(nsel+1,status.MPI_SOURCE*factor_+1); assert(add>0);
      while( pb-buff < tt ) {
	todo--;            assert(todo>=0); 
	int lid = *pb++, mina = *pb++, nc = *pb++; 
	assert(lid < nNodes_ && lid >= nLocalNd_);
	curr = nodes_[lid];
	assert(curr->nChildInSel() == 0); 
	ierr = curr->SetnChildInSel( nc );  CHKERRQ(ierr);  
	assert(curr->minAdjInSel() == 0); 
	ierr = curr->SetMinAdjInSel( mina );  CHKERRQ(ierr);  
      }
      assert(pb-buff == tt);      PetscFree(buff);
    }
    MPI_Iprobe( MPI_ANY_SOURCE, tag, gComm_, &ii, &status ); assert(!ii);
    // cleanup sends
    while( !smessages.isEmpty() ) {
      MPI_Request *sreq = smessages.removeHead();
      MPI_Test( sreq, &ii, &status );
      if( ii ) { MPI_Wait( sreq, &status );    PetscFree( sreq ); }
      else smessages.addTail( sreq );
    }
  }
  assert( proc_list.isEmpty() && smessages.isEmpty() );
  
  return 0;
}
  
#undef __FUNCT__    
#define __FUNCT__ "PromAggGrid::AddDepsToList" 
/* PromAggGrid::AddDepsToList ******************************************
 *
 *  - 
 *
 *  INPUT:
 *    - procTable: (in/out) 
 *
 *  SIDE EFFECTS:
 *    - 
 *
 *  RETURN:
 *    - PET cerror code
 *
 */
int PromAggGrid::AddDepsToList( PromTable<int> &procTable )
{
  PromNode *curr;
  Position pos;
  PromGlobalID *dep;

  // procTable -- for next
  for( int xx = 0 ; xx < nNodes_ ; xx++ ) {
    curr = nodes_[xx];
    if( curr->isGhost() ) procTable.add(curr->symbData_->ghostID.proc()+1); 
    if( curr->symbData_->dependents ) {
      pos = curr->symbData_->dependents->getHeadPosition();assert(pos!=NULL);
      while( pos ) { // send to old ghosts
	dep = curr->symbData_->dependents->getNext(pos);   
	procTable.add(dep->proc()+1); 
      }
    }
  }
  
  return 0;
}

/* PromAggGrid::PickSubdomains ********************************************
 *
 *  - completes data structures after METIS partitions to mimic MIS 
 *
 *  INPUT:
 *    - pnode_part:
 *
 *  SIDE EFFECTS:
 *    - selectes node for promotion and constructs child list for selected
 *
 *  RETURN:
 *    - PET cerror code 
 */
#undef __FUNCT__ 
#define __FUNCT__ "PromAggGrid::PickSubdomains" 
int PromAggGrid::PickSubdomains( const int pnode_part[] )
{
  int        xx;
  PromList<PromNode*> list1, list2, list3;
  PromNode *curr, *c2; 
  
  // pick local partitions  
  for( xx = 0 ; xx < nNodes_ ; xx++ ) nodes_[xx]->kill(FALSE); 
  for( xx = 0; xx < nLocalNd_ ; xx++ ) { // local nodes only
    curr = nodes_[xx]; 
    if( curr->isDead() ) continue; 
    // breath first search 
    int parti = pnode_part[xx], nn = 0;
    curr->kill(); nn++; list3.addTail( curr );
    list1.addTail( curr );
    do{
      while( !list1.isEmpty() ) {
	c2 = list1.removeHead();
	PromTable<PromNode*> *adj = &c2->symbData_->adjac;
	TablePos tpos = adj->getHeadPosition(); 
	while( tpos ){ 
	  PromNode *c3 = adj->getNext(tpos);
	  if( c3->isGhost() ) continue;
	  if( !c3->isDead() && pnode_part[c3->index_()] == parti ) {
	    list2.addTail( c3 ); c3->kill(); nn++; list3.addTail(c3);
	  }
	}
      }
      while( !list2.isEmpty() ) list1.addTail( list2.removeHead() );
    }while( !list1.isEmpty() );
    assert( nn > 0 ); assert( nn == list3.getCount() );
    
    // get "root" on boundary (minimize new ghosts next door)
    curr = NULL;    
    Position pos = list3.getHeadPosition(); 
    while( pos ) {
      c2 = list3.getNext( pos );   
      if( !curr || 
	  (c2->isBound() && (!curr->isBound() ||
			     c2->symbData_->dependents->getCount() >
			     curr->symbData_->dependents->getCount())))curr=c2;
    }
    assert(curr!=NULL); // this is (arbitrarily) promoted
    curr->setState( PROM_SELECTED ); assert(curr->symbData_->parent == NULL);
    while( !list3.isEmpty() ) {
      c2 = list3.removeHead();
      if( c2 != curr ) {        // set lists
	curr->addChild( c2 );	c2->setState( PROM_DELETED ); 
	c2->symbData_->parent = curr;
      }
    }
  }
  // clean markes
  for( xx = 0 ; xx < nNodes_ ; xx++ ) {
    curr = nodes_[xx];  curr->kill(FALSE); 
  }

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromAggGrid::SetAMGEdgeWeights" 
/* PromAggGrid::SetAMGEdgeWeights *******************************************
 *
 *  - PETSc
 *
 *  INPUT: 
 *    - AA:
 *    - adjncy[numadjac]: (out)
 *    - numadjacTOT:
 *    - egwtsTOT:
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code
 */ 
int PromAggGrid::SetAMGEdgeWeights( PromMatrix *AA, idxtype *adjncy,
				    const int numadjacTOT, int *egwtsTOT )
{
  int *arrp,ii,wht,xx,ti,tt,maxw,proc,ncols,ierr,todo,actE,adjacs[PROM_MAXNIEGH];
  double *naiiArr, naij,naii,najj; double sumw=0.0,nwht=0.0;
  double *Ai[PROM_MAXNIEGH]; const int tag=getNewTag(),tag1=getNewTag();
  MPI_Status status; assert(isActive());
  PromGlobalID *dep;
  PromList<int> *list;           xx = sizeof(MPI_Request); ii = sizeof(int);
  PromTable<PromList<int>*> proc_list(57);    tt = sizeof(double);
  const int fisz = tt/ii, reqsz = (xx/tt + ((xx%tt == 0) ? 0 : 1)) * fisz;
  const int my0 = proc_gnode_[comm_.mype()], fact=AA->isComplex() ? 2 : 1;
  PromList<MPI_Request*> smessages;
  PromTable<int> *gid_wht;

  // table array
  ii = prom_table_size; prom_table_size = 0;
  PromTable<int> * const lid_gid_wht = new PromTable<int>[nLocalNd_+1];
  prom_table_size = ii;
  ierr = PetscMalloc( 2*prom_table_size*(nLocalNd_+1)*sizeof(int), &arrp);  
  CHKERRQ(ierr);
  for( ii = tt = 0, maxw = prom_table_size, xx = nLocalNd_*maxw ; 
       ii < nLocalNd_ ; 
       ii++, tt += maxw, xx += maxw ){
    lid_gid_wht[ii].SetTable( &arrp[tt], maxw ); CHKERRQ(ierr);
    lid_gid_wht[ii].SetKeyTable( &arrp[xx], maxw ); CHKERRQ(ierr);
  }
  // get diag (A_ii), collect sends for diag and A_ij, j > i
  ierr = PetscMalloc((nNodes_+1)*sizeof(double),&naiiArr); CHKERRQ(ierr);
  naiiArr[nNodes_] = -22.;
  for( xx = 0 ; xx < nLocalNd_ ; xx++ ) { 
    PromNode *curr = nodes_[xx];
    ierr = AA->GetLocalNodeRow( xx, PROM_MAXNIEGH, ncols, adjacs, Ai ); 
    CHKERRQ(ierr);
    if(ncols==0)PromContext::printArr(stderr,adjacs,ncols,xx,"zero row rowIDs[%d] row %d");
    const int indf = curr->getNDF();
    ierr = get_norm(indf, fact, curr->index_.global(), indf, ncols, adjacs, Ai, &naij); CHKERRQ(ierr);
    naiiArr[xx] = naij;
    if( curr->isBound() ){
      Position pos = curr->symbData_->dependents->getHeadPosition(); 
      while( pos ) {
	dep = curr->symbData_->dependents->getNext(pos); 
	proc = dep->proc(); 
	if( (list=proc_list.find(proc+1)) == NULL ) {
	  list = new PromList<int>(100); CHKNEW(list,1);
	  proc_list.add( list, proc+1 );
	}
	list->addTail( xx+1 ); 
	list->addTail( dep->local() + 1 ); 
      }
    }
  } 

  // send diag
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--;
    tt = list->getCount(); assert(tt%2==0); tt = (tt/2) * ( 2 + fisz ); 
    MPI_Request *sreq; 
    ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); CHKERRQ(ierr);
    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    while( !list->isEmpty() ) {
      xx = list->removeHead() - 1; 
      int deploc = list->removeHead() - 1; 
      *pb++ = deploc; *pb++ = xx; double *fp=(double*)pb; *fp++ = naiiArr[xx];
      pb = (int*)fp;
    }
    assert(tt == pb-buff);
    // send(tag) *< lid, xx, Aii > 
    MPI_Isend( buff, tt, MPI_INT, proc/factor_, tag, gComm_, sreq );
    delete list;
  }
  // recv diag
  ii = nNodes_ - nLocalNd_; 
  while( ii > 0 ) {
    MPI_Probe( MPI_ANY_SOURCE, tag, gComm_, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); 
    int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb=buff;
    CHKERRQ(ierr);
    // receive(tag) *< lid, xx, Aii >
    MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag, gComm_, &status);
    while( pb-buff < tt ) {
      int lid = *pb++; xx = *pb++; 
      double *fp = (double*)pb; naiiArr[lid] = *fp++; pb = (int*)fp;
      ii--;
    }
    assert(pb-buff == tt);    PetscFree(buff);
  }
  assert(ii==0);
  // cleanup sends 
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  MPI_Iprobe(MPI_ANY_SOURCE, tag, gComm_, &ii, &status);assert(!ii); 
  // collect and set j > i 
  for( xx = ti = ii = maxw = todo = 0 ; xx < nLocalNd_ ; xx++ ) { 
    PromNode *curr = nodes_[xx];
    naii = naiiArr[xx]; 
    ierr = AA->GetLocalNodeRow(xx,PROM_MAXNIEGH, ncols, adjacs, Ai); 
    CHKERRQ(ierr);
    const int indf = curr->getNDF(), gidi = curr->index_.global();
    // keep same order in 'adjncy' as node lists, skip regulaization
    TablePos tpos = curr->symbData_->adjac.getHeadPosition();
    while( tpos ) { 
      PromNode *n1 = curr->symbData_->adjac.getNext(tpos);
      if( n1->getNDF() == curr->getNDF() ) {
	const int gidj = n1->index_.global(), lidj = n1->index_();
	if( gidj < gidi ) { 
	  wht = -12; 
	  if( n1->isGhost() ) todo++; // will recv. this
	} 
	else { // will need to add with trans. (recv)
	  assert( gidj > gidi );
	  ierr = get_norm(indf, fact, gidj, indf, ncols, adjacs, Ai, &naij); 
	  najj = naiiArr[lidj];
	  get_weight( naij, naii, najj, &wht ); assert(wht>0); 
	  sumw += (double)wht; nwht += 1.0; 
	  // put transpose
	  if( !n1->isGhost() ) { // local tanspose
	    gid_wht = &lid_gid_wht[lidj];
	    int add = gid_wht->add( wht+1, gidi+1 ); assert(add==1);
	  }
	  else { // collect to send
	    proc = n1->symbData_->ghostID.proc();
	    if( (list=proc_list.find(proc+1)) == NULL ) {
	      list = new PromList<int>(100); CHKNEW(list,1);
	      proc_list.add( list, proc+1 );
	    }
	    list->addTail( curr->index_.global() + 1 ); 
	    list->addTail( n1->index_.global() + 1 ); 
	    list->addTail( wht+1 ); // hack
	  }
	}
	if( wht > maxw ) maxw = wht;
	egwtsTOT[ti++] = wht;
	if( !global_agg ) {
	  if( !n1->isGhost() ) {
	    ewgts[ii] = wht; adjncy[ii++] = lidj; // "adjncy" folded in here 
	  } 
	}
	else {
	  ewgts[ii] = wht; adjncy[ii++] = gidj; // "adjncy" folded in here 
	}
      }
    }
  }
  if(ii!=numadjac)SETERRQ2(1,"ii=%d numadjac=%d",ii,numadjac);
  assert(ii==numadjac);

  // send tanspose (j > i)
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--;
    tt = list->getCount();    assert(tt%3==0); 
    MPI_Request *sreq; 
    ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); CHKERRQ(ierr);
    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    while( !list->isEmpty() ) {
      *pb++ = list->removeHead() - 1; // gidi
      *pb++ = list->removeHead() - 1; // gidj
      *pb++ = list->removeHead() - 1; // wht
    }
    assert(tt == pb-buff);
    // send(tag1) *< glidi, gidj, wht > 
    MPI_Isend( buff, tt, MPI_INT, proc/factor_, tag1, gComm_, sreq );
    delete list; 
  }
  // recv tanspose (i > j)
  while( todo ) {
    MPI_Probe( MPI_ANY_SOURCE, tag1, gComm_, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); 
    int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb=buff;
    CHKERRQ(ierr);
    // receive(tag1) *< glidi, gidj, wht >
    MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag1, gComm_, &status);
    while( pb-buff < tt ) {
      int gidi = *pb++, gidj = *pb++, wht =  *pb++, lidj = gidj - my0;
      gid_wht = &lid_gid_wht[lidj]; assert(wht>=0); 
      int add = gid_wht->add( wht+1, gidi+1 ); assert(add==1);
      todo--;
    }
    assert(pb-buff == tt);    PetscFree(buff);
  }
  // set i > j
  for( xx = ti = ii = actE = 0 ; xx < nLocalNd_ ; xx++ ) { 
    PromNode *curr = nodes_[xx];
    const int gidi = curr->index_.global(), lidi = curr->index_();
    gid_wht = &lid_gid_wht[lidi]; 
    // keep same order in 'adjncy' as node lists, skip regulaization
    TablePos tpos = curr->symbData_->adjac.getHeadPosition();
    while( tpos ) { 
      PromNode *n1 = curr->symbData_->adjac.getNext(tpos);
      if( n1->getNDF() == curr->getNDF() ) {
	const int gidj = n1->index_.global();
	if( gidj < gidi ) { 
	  wht = gid_wht->find( gidj+1 ) - 1;  assert(wht>=0);
	} 
	else wht = egwtsTOT[ti]; // just set again for j > i
	if( wht > maxw ) maxw = wht;
	egwtsTOT[ti++] = wht;
	if( !global_agg ) {
	  if( n1->isGhost() ) continue;	    
	  ewgts[ii] = wht;
	  if( wht > edgeW_limit ) actE++;
	  adjncy[ii++] = n1->index_(); // "adjncy" folded in here 
	} 
	else {
	  ewgts[ii] = wht;
	  if( wht > edgeW_limit ) actE++;
	  adjncy[ii++] = n1->index_.global(); // "adjncy" folded in here 
	}
      }
    }
  } 
  // cleanup sends 
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  // cleanup
  assert(naiiArr[nNodes_] == -22.);
  assert(numadjac == ii);  assert(ti == numadjacTOT);
  PetscFree(naiiArr);
  delete [] lid_gid_wht;
  PetscFree(arrp);

  {
    double sarr[5]={sumw,nwht,(double)maxw,(double)actE,(double)numadjac};
    double arr[5];
    MPI_Allreduce( sarr, arr, 5, MPI_DOUBLE, MPI_SUM, gComm_ );
    double g_avewht = arr[0]/(arr[1]+1.e-200),gActE=arr[3],gNAdj=arr[4];
    if( (double)gActE/gNAdj < 0.2 && edgeW_limit > EMAX/200 &&
	gActE/(double)AA->map_.nglobal() < 8.0 ) {
      edgeW_limit = edgeW_limit/2;
      if( options_.verbose_ > 0 ) {
	PetscPrintf(gComm_,"\t[%d] retry weights %d, %g edges active\n",
		    comm_.mype(),edgeW_limit,(double)gActE/gNAdj);
      }
      //
      ierr = SetAMGEdgeWeights(AA,adjncy,numadjacTOT,egwtsTOT); CHKERRQ(ierr);
    }
    else if(gActE/(double)AA->map_.nglobal() < 2.0 ||  // too small
	    (gActE/(double)AA->map_.nglobal()<4.0 && (double)gActE/gNAdj<0.2)){
      if( options_.verbose_ > 0 ) {
	PetscPrintf(gComm_,
		    "\t[%d] Deleteing weights. weights limit=%d ave=%d, %g%% active edges, %g ave. edge/node\n",
		    comm_.mype(),edgeW_limit,(int)g_avewht,
		    100.0 * (double)actE/(double)(numadjac+1),
		    gActE/(double)AA->map_.nglobal());
      }
      PetscFree(ewgts); ewgts = NULL;
    }
    else if( options_.verbose_ > 1 ) {
      PetscPrintf(gComm_,
		  "[%d] weights limit=%d ave=%d, %g%% active edges, %g ave. edge/node\n",
		  comm_.mype(),edgeW_limit,(int)g_avewht,
		  100.0 * (double)actE/(double)(numadjac+1),
		  gActE/(double)AA->map_.nglobal() );
    }
  }
  
  return 0;
}
